/*
    libmaus2
    Copyright (C) 2020 German Tischler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_NETWORK_CURLOUTPUTBLOCK_HPP)
#define LIBMAUS2_NETWORK_CURLOUTPUTBLOCK_HPP

#include <libmaus2/util/AutoArrayCharBaseAllocator.hpp>
#include <libmaus2/util/AutoArrayCharBaseTypeInfo.hpp>

struct CurlOutputBlock
{
	libmaus2::util::AutoArrayCharBaseTypeInfo::pointer_type aptr;
	char const * a;
	char const * e;

	CurlOutputBlock() : aptr(), a(nullptr), e(nullptr) {}
	CurlOutputBlock(
		libmaus2::util::AutoArrayCharBaseTypeInfo::pointer_type raptr,
		::std::size_t const n
	);
};
#endif
