/*
    libmaus2
    Copyright (C) 2015 German Tischler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#undef LIBMAUS2_LCS_SIMD_BANDED_CLASS_NAME
#undef LIBMAUS2_LCS_SIMD_BANDED_WORD_TYPE
#undef LIBMAUS2_LCS_SIMD_BANDED_ELEMENT_TYPE
#undef LIBMAUS2_LCS_SIMD_BANDED_FIRST_FF_REST_0
#undef LIBMAUS2_LCS_SIMD_BANDED_LAST_FF_REST_0
#undef LIBMAUS2_LCS_SIMD_BANDED_ALL_ONE
#undef LIBMAUS2_LCS_SIMD_BANDED_ALL_0
#undef LIBMAUS2_LCS_SIMD_BANDED_ALL_1
#undef LIBMAUS2_LCS_SIMD_BANDED_SELECT_FIRST_TOBACK
#undef LIBMAUS2_LCS_SIMD_BANDED_STORE
#undef LIBMAUS2_LCS_SIMD_BANDED_LOAD_ALIGNED
#undef LIBMAUS2_LCS_SIMD_BANDED_LOAD_UNALIGNED
#undef LIBMAUS2_LCS_SIMD_BANDED_ANDNOT
#undef LIBMAUS2_LCS_SIMD_BANDED_AND
#undef LIBMAUS2_LCS_SIMD_BANDED_OR
#undef LIBMAUS2_LCS_SIMD_BANDED_CMPEQ
#undef LIBMAUS2_LCS_SIMD_BANDED_ADD
#undef LIBMAUS2_LCS_SIMD_BANDED_SHUFFLE
#undef LIBMAUS2_LCS_SIMD_BANDED_MIN
#undef LIBMAUS2_LCS_SIMD_BANDED_SHIFTOVER
#undef LIBMAUS2_LCS_SIMD_BANDED_SHIFTOVERLEFT
#undef LIBMAUS2_LCS_SIMD_BANDED_INIT
#undef LIBMAUS2_LCS_SIMD_BANDED_SHIFTRIGHT
#undef LIBMAUS2_LCS_SIMD_BANDED_SELECTLAST
#undef LIBMAUS2_LCS_SIMD_BANDED_SHIFTLEFT
#undef LIBMAUS2_LCS_SIMD_BANDED_FIRST_TO_BACK
