/*
    libmaus2
    Copyright (C) 2015 German Tischler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_DAZZLER_ALIGN_OVERLAPHEAPCOMPARATOR_HPP)
#define LIBMAUS2_DAZZLER_ALIGN_OVERLAPHEAPCOMPARATOR_HPP

#include <libmaus2/dazzler/align/Overlap.hpp>

namespace libmaus2
{
	namespace dazzler
	{
		namespace align
		{
			template<typename comparator_type>
			struct OverlapHeapComparator
			{
				comparator_type comparator;

				OverlapHeapComparator(comparator_type rcomparator = comparator_type())
				: comparator(rcomparator) {}

				bool operator()(std::pair<uint64_t,libmaus2::dazzler::align::Overlap> const & A, std::pair<uint64_t,libmaus2::dazzler::align::Overlap> const & B) const
				{
					// B < A -> A > B
					if ( comparator(B.second,A.second) )
						return true;
					// A < B -> B > A
					else if ( comparator(A.second,B.second) )
						return false;
					// A == B, compare by id
					else
						return A.first > B.first;
				}
			};
		}
	}
}
#endif
