/*
    libmaus2
    Copyright (C) 2016 German Tischler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_GAMMA_INTERVAL_HPP)
#define LIBMAUS2_GAMMA_INTERVAL_HPP

#include <libmaus2/types/types.hpp>

namespace libmaus2
{
	namespace gamma
	{
		struct Interval
		{
			uint64_t from;
			uint64_t to;

			Interval() : from(0), to(0)
			{

			}

			Interval(uint64_t const rfrom, uint64_t const rto)
			: from(rfrom), to(rto)
			{

			}

			bool operator==(Interval const & P) const
			{
				return
					from == P.from && to == P.to;
			}
		};
	}
}
#endif
