/**
    libmaus2
    Copyright (C) 2009-2016 German Tischler
    Copyright (C) 2011-2014 Genome Research Limited

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#if ! defined(LIBMAUS2_SUFFIXSORT_BWTB3MBASE_HPP)
#define LIBMAUS2_SUFFIXSORT_BWTB3MBASE_HPP

#include <libmaus2/huffman/RLEncoder.hpp>
#include <libmaus2/huffman/RLDecoder.hpp>
#include <libmaus2/gamma/GammaRLEncoder.hpp>
#include <libmaus2/gamma/GammaRLDecoder.hpp>

namespace libmaus2
{
	namespace suffixsort
	{
		namespace bwtb3m
		{
			struct BWTB3MBase
			{
				// #define LIBMAUS2_SUFFIXSORT_BWTB3M_HUFGAP
				#define LIBMAUS2_SUFFIXSORT_BWTB3M_HUFRL

				#if defined(LIBMAUS2_SUFFIXSORT_BWTB3M_HUFRL)
				typedef ::libmaus2::huffman::RLEncoderStd rl_encoder;
				typedef ::libmaus2::huffman::RLDecoder rl_decoder;
				#else
				typedef ::libmaus2::gamma::GammaRLEncoder rl_encoder;
				typedef ::libmaus2::gamma::GammaRLDecoder rl_decoder;
				#endif
			};
		}
	}
}
#endif
