/*
    libmaus2
    Copyright (C) 2009-2014 German Tischler
    Copyright (C) 2011-2014 Genome Research Limited

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_DIGEST_DIGESTS_HPP)
#define LIBMAUS2_DIGEST_DIGESTS_HPP

#include <libmaus2/digest/md5.hpp>
#include <libmaus2/digest/CRC32.hpp>
#include <libmaus2/digest/CRC32C.hpp>
#include <libmaus2/digest/Null.hpp>
#include <libmaus2/digest/SHA1.hpp>
#include <libmaus2/digest/SHA2_224.hpp>
#include <libmaus2/digest/SHA2_256.hpp>
#include <libmaus2/digest/SHA2_384.hpp>
#include <libmaus2/digest/SHA2_512.hpp>
#include <libmaus2/digest/MurmurHash3_x64_128.hpp>
#include <libmaus2/digest/SHA3_256.hpp>
#endif
