/*
    libmaus2
    Copyright (C) 2018 German Tischler-Höhle

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if !defined(LIBMAUS2_AIO_INPUTSTREAMOBJECTALLOCATOR_HPP)
#define LIBMAUS2_AIO_INPUTSTREAMOBJECTALLOCATOR_HPP

#include <libmaus2/aio/InputStreamObject.hpp>

namespace libmaus2
{
	namespace aio
	{
		struct InputStreamObjectAllocator
		{
			std::istream * istr;
			uint64_t streamid;

			InputStreamObjectAllocator() {}
			InputStreamObjectAllocator(std::istream * ristr, uint64_t const rstreamid) : istr(ristr), streamid(rstreamid) {}

			InputStreamObject::shared_ptr_type operator()()
			{
				InputStreamObject::shared_ptr_type tptr(new InputStreamObject(istr,streamid));
				return tptr;
			}
		};
	}
}
#endif
