/*
    libmaus2
    Copyright (C) 2009-2015 German Tischler
    Copyright (C) 2011-2015 Genome Research Limited

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_FASTX_REFPATHTOKEN_HPP)
#define LIBMAUS2_FASTX_REFPATHTOKEN_HPP

#include <ostream>
#include <string>

namespace libmaus2
{
	namespace fastx
	{
		struct RefPathToken
		{
			enum path_token { path_token_literal, path_token_s };

			path_token token;
			std::string s;
			size_t n;

			RefPathToken() : token(path_token_literal), s(), n(0) {}
			RefPathToken(std::string const & rs) : token(path_token_literal), s(rs), n(0) {}
			RefPathToken(size_t const rn) : token(path_token_s), s(), n(rn) {}
		};

		std::ostream & operator<<(std::ostream & out, RefPathToken const & O);
	}
}
#endif
