/*
    libmaus2
    Copyright (C) 2020 German Tischler-Höhle

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#if ! defined(LIBMAUS2_PARALLEL_THREADPOOL_BGZF_BGZFDECOMPRESSPACKAGEDATADISPATCHER_HPP)
#define LIBMAUS2_PARALLEL_THREADPOOL_BGZF_BGZFDECOMPRESSPACKAGEDATADISPATCHER_HPP

#include <libmaus2/parallel/threadpool/bgzf/BgzfDecompressPackageData.hpp>
#include <libmaus2/parallel/threadpool/bgzf/BgzfBlockInfoEnqueDecompressHandler.hpp>

namespace libmaus2
{
	namespace parallel
	{
		namespace threadpool
		{
			namespace bgzf
			{
				struct BgzfDecompressPackageDataDispatcher : public libmaus2::parallel::threadpool::ThreadPoolDispatcher
				{
					virtual void dispatch(
						libmaus2::util::shared_ptr<libmaus2::parallel::threadpool::ThreadWorkPackage> package,
						libmaus2::parallel::threadpool::ThreadPool * /* pool */
					)
					{
						BgzfDecompressPackageData * data = dynamic_cast<BgzfDecompressPackageData *>(package->data.load().get());
						BgzfBlockInfoEnqueDecompressHandler * handler = data->decompHandler;
						handler->decompressData(package->data);
					}
				};
			}
		}
	}
}
#endif
